@@ -9,10 +9,13 @@
 
 #include "adw-application-window.h"
 
+#include "adw-adaptive-preview-private.h"
 #include "adw-breakpoint-bin-private.h"
 #include "adw-dialog-host-private.h"
 #include "adw-dialog-private.h"
 #include "adw-gizmo-private.h"
+#include "adw-main-private.h"
+#include "adw-widget-utils-private.h"
 
 /**
  * AdwApplicationWindow:
@@ -55,6 +58,7 @@ typedef struct
   GtkWidget *titlebar;
   GtkWidget *bin;
   GtkWidget *dialog_host;
+  GtkWidget *adaptive_preview;
 } AdwApplicationWindowPrivate;
 
 static void adw_application_window_buildable_init (GtkBuildableIface *iface);
@@ -71,6 +75,7 @@ enum {
   PROP_CURRENT_BREAKPOINT,
   PROP_DIALOGS,
   PROP_VISIBLE_DIALOG,
+  PROP_ADAPTIVE_PREVIEW,
   LAST_PROP,
 };
 
@@ -88,6 +93,25 @@ notify_visible_dialog_cb (AdwApplicationWindow *self)
   g_object_notify_by_pspec (G_OBJECT (self), props[PROP_VISIBLE_DIALOG]);
 }
 
+static void
+adaptive_preview_exit_cb (AdwApplicationWindow *self)
+{
+  adw_application_window_set_adaptive_preview (self, FALSE);
+}
+
+static gboolean
+toggle_adaptive_preview_cb (AdwApplicationWindow *self)
+{
+  if (!adw_get_inspector_keybinding_enabled ())
+    return GDK_EVENT_PROPAGATE;
+
+  gboolean open = adw_application_window_get_adaptive_preview (self);
+
+  adw_application_window_set_adaptive_preview (self, !open);
+
+  return GDK_EVENT_STOP;
+}
+
 static void
 adw_application_window_size_allocate (GtkWidget *widget,
                                       int        width,
@@ -96,12 +120,14 @@ adw_application_window_size_allocate (GtkWidget *widget,
 {
   AdwApplicationWindow *self = ADW_APPLICATION_WINDOW (widget);
   AdwApplicationWindowPrivate *priv = adw_application_window_get_instance_private (self);
+  GtkWidget *child;
 
   /* We don't want to allow any other titlebar */
   if (gtk_window_get_titlebar (GTK_WINDOW (self)) != priv->titlebar)
     g_error ("gtk_window_set_titlebar() is not supported for AdwApplicationWindow");
 
-  if (gtk_window_get_child (GTK_WINDOW (self)) != priv->dialog_host)
+  child = gtk_window_get_child (GTK_WINDOW (self));
+  if (child != priv->dialog_host && child != priv->adaptive_preview)
     g_error ("gtk_window_set_child() is not supported for AdwApplicationWindow");
 
   GTK_WIDGET_CLASS (adw_application_window_parent_class)->size_allocate (widget,
@@ -131,6 +157,9 @@ adw_application_window_get_property (GObject    *object,
   case PROP_VISIBLE_DIALOG:
     g_value_set_object (value, adw_application_window_get_visible_dialog (self));
     break;
+  case PROP_ADAPTIVE_PREVIEW:
+    g_value_set_boolean (value, adw_application_window_get_adaptive_preview (self));
+    break;
   default:
     G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
   }
@@ -148,6 +177,9 @@ adw_application_window_set_property (GObject      *object,
   case PROP_CONTENT:
     adw_application_window_set_content (self, g_value_get_object (value));
     break;
+  case PROP_ADAPTIVE_PREVIEW:
+    adw_application_window_set_adaptive_preview (self, g_value_get_boolean (value));
+    break;
   default:
     G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
   }
@@ -164,7 +196,7 @@ adw_application_window_class_init (AdwApplicationWindowClass *klass)
   widget_class->size_allocate = adw_application_window_size_allocate;
 
   /**
-   * AdwApplicationWindow:content: (attributes org.gtk.Property.get=adw_application_window_get_content org.gtk.Property.set=adw_application_window_set_content)
+   * AdwApplicationWindow:content:
    *
    * The content widget.
    *
@@ -176,7 +208,7 @@ adw_application_window_class_init (AdwApplicationWindowClass *klass)
                          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_EXPLICIT_NOTIFY);
 
   /**
-   * AdwApplicationWindow:current-breakpoint: (attributes org.gtk.Property.get=adw_application_window_get_current_breakpoint)
+   * AdwApplicationWindow:current-breakpoint:
    *
    * The current breakpoint.
    *
@@ -188,7 +220,7 @@ adw_application_window_class_init (AdwApplicationWindowClass *klass)
                          G_PARAM_READABLE | G_PARAM_STATIC_STRINGS);
 
   /**
-   * AdwApplicationWindow:dialogs: (attributes org.gtk.Property.get=adw_application_window_get_dialogs)
+   * AdwApplicationWindow:dialogs:
    *
    * The open dialogs.
    *
@@ -200,7 +232,7 @@ adw_application_window_class_init (AdwApplicationWindowClass *klass)
                          G_PARAM_READABLE | G_PARAM_STATIC_STRINGS);
 
   /**
-   * AdwApplicationWindow:visible-dialog: (attributes org.gtk.Property.get=adw_application_window_get_visible_dialog)
+   * AdwApplicationWindow:visible-dialog:
    *
    * The currently visible dialog
    *
@@ -211,6 +243,26 @@ adw_application_window_class_init (AdwApplicationWindowClass *klass)
                          ADW_TYPE_DIALOG,
                          G_PARAM_READABLE | G_PARAM_STATIC_STRINGS);
 
+  /**
+   * AdwApplicationWindow:adaptive-preview:
+   *
+   * Whether adaptive preview is currently open.
+   *
+   * Adaptive preview is a debugging tool used for testing the window
+   * contents at specific screen sizes, simulating mobile environment.
+   *
+   * Adaptive preview can always be accessed from inspector. This function
+   * allows applications to open it manually.
+   *
+   * Most applications should not use this property.
+   *
+   * Since: 1.7
+   */
+  props[PROP_ADAPTIVE_PREVIEW] =
+    g_param_spec_boolean ("adaptive-preview", NULL, NULL,
+                          FALSE,
+                          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_EXPLICIT_NOTIFY);
+
   g_object_class_install_properties (object_class, LAST_PROP, props);
 }
 
@@ -218,6 +270,8 @@ static void
 adw_application_window_init (AdwApplicationWindow *self)
 {
   AdwApplicationWindowPrivate *priv = adw_application_window_get_instance_private (self);
+  GtkEventController *shortcut_controller;
+  GtkShortcut *shortcut;
 
   priv->titlebar = adw_gizmo_new_with_role ("nothing", GTK_ACCESSIBLE_ROLE_PRESENTATION,
                                             NULL, NULL, NULL, NULL, NULL, NULL);
@@ -238,7 +292,20 @@ adw_application_window_init (AdwApplicationWindow *self)
   g_signal_connect_swapped (priv->dialog_host, "notify::visible-dialog",
                             G_CALLBACK (notify_visible_dialog_cb), self);
 
+  gtk_widget_set_size_request (GTK_WIDGET (self), 360, 200);
+
   gtk_application_window_set_show_menubar (GTK_APPLICATION_WINDOW (self), FALSE);
+
+  if (adw_is_adaptive_preview ())
+    adw_application_window_set_adaptive_preview (self, TRUE);
+
+  shortcut = gtk_shortcut_new (gtk_keyval_trigger_new (GDK_KEY_M, GDK_CONTROL_MASK | GDK_SHIFT_MASK),
+                               gtk_callback_action_new ((GtkShortcutFunc) toggle_adaptive_preview_cb, self, NULL));
+
+  shortcut_controller = gtk_shortcut_controller_new ();
+  gtk_shortcut_controller_set_scope (GTK_SHORTCUT_CONTROLLER (shortcut_controller), GTK_SHORTCUT_SCOPE_GLOBAL);
+  gtk_shortcut_controller_add_shortcut (GTK_SHORTCUT_CONTROLLER (shortcut_controller), shortcut);
+  gtk_widget_add_controller (GTK_WIDGET (self), shortcut_controller);
 }
 
 static void
@@ -283,7 +350,7 @@ adw_application_window_new (GtkApplication *app)
 }
 
 /**
- * adw_application_window_set_content: (attributes org.gtk.Method.set_property=content)
+ * adw_application_window_set_content:
  * @self: an application window
  * @content: (nullable): the content widget
  *
@@ -300,21 +367,21 @@ adw_application_window_set_content (AdwApplicationWindow *self,
   g_return_if_fail (ADW_IS_APPLICATION_WINDOW (self));
   g_return_if_fail (content == NULL || GTK_IS_WIDGET (content));
 
-  if (content)
-    g_return_if_fail (gtk_widget_get_parent (content) == NULL);
-
   priv = adw_application_window_get_instance_private (self);
 
   if (adw_application_window_get_content (self) == content)
     return;
 
+  if (content)
+    g_return_if_fail (gtk_widget_get_parent (content) == NULL);
+
   adw_breakpoint_bin_set_child (ADW_BREAKPOINT_BIN (priv->bin), content);
 
   g_object_notify_by_pspec (G_OBJECT (self), props[PROP_CONTENT]);
 }
 
 /**
- * adw_application_window_get_content: (attributes org.gtk.Method.get_property=content)
+ * adw_application_window_get_content:
  * @self: an application window
  *
  * Gets the content widget of @self.
@@ -359,7 +426,7 @@ adw_application_window_add_breakpoint (AdwApplicationWindow *self,
 }
 
 /**
- * adw_application_window_get_current_breakpoint: (attributes org.gtk.Method.get_property=current-breakpoint)
+ * adw_application_window_get_current_breakpoint:
  * @self: an application window
  *
  * Gets the current breakpoint.
@@ -381,7 +448,7 @@ adw_application_window_get_current_breakpoint (AdwApplicationWindow *self)
 }
 
 /**
- * adw_application_window_get_dialogs: (attributes org.gtk.Method.get_property=dialogs)
+ * adw_application_window_get_dialogs:
  * @self: an application window
  *
  * Returns a [iface@Gio.ListModel] that contains the open dialogs of @self.
@@ -405,7 +472,7 @@ adw_application_window_get_dialogs (AdwApplicationWindow *self)
 }
 
 /**
- * adw_application_window_get_visible_dialog: (attributes org.gtk.Method.get_property=visible-dialog)
+ * adw_application_window_get_visible_dialog:
  * @self: an application window
  *
  * Returns the currently visible dialog in @self, if there's one.
@@ -425,3 +492,76 @@ adw_application_window_get_visible_dialog (AdwApplicationWindow *self)
 
   return adw_dialog_host_get_visible_dialog (ADW_DIALOG_HOST (priv->dialog_host));
 }
+
+/**
+ * adw_application_window_get_adaptive_preview:
+ * @self: an application window
+ *
+ * Gets whether adaptive preview for @self is currently open.
+ *
+ * Returns: whether adaptive preview is open.
+ *
+ * Since: 1.7
+ */
+gboolean
+adw_application_window_get_adaptive_preview (AdwApplicationWindow *self)
+{
+  AdwApplicationWindowPrivate *priv;
+
+  g_return_val_if_fail (ADW_IS_APPLICATION_WINDOW (self), FALSE);
+
+  priv = adw_application_window_get_instance_private (self);
+
+  return priv->adaptive_preview != NULL;
+}
+
+/**
+ * adw_application_window_set_adaptive_preview:
+ * @self: an application window
+ * @adaptive_preview: whether to open adaptive preview
+ *
+ * Sets whether adaptive preview for @self is currently open.
+ *
+ * Adaptive preview is a debugging tool used for testing the window
+ * contents at specific screen sizes, simulating mobile environment.
+ *
+ * Adaptive preview can always be accessed from inspector. This function
+ * allows applications to open it manually.
+ *
+ * Most applications should not use this function.
+ *
+ * Since: 1.7
+ */
+void
+adw_application_window_set_adaptive_preview (AdwApplicationWindow *self,
+                                             gboolean              adaptive_preview)
+{
+  AdwApplicationWindowPrivate *priv;
+
+  g_return_if_fail (ADW_IS_APPLICATION_WINDOW (self));
+
+  priv = adw_application_window_get_instance_private (self);
+
+  if (adaptive_preview == adw_application_window_get_adaptive_preview (self))
+    return;
+
+  g_object_ref (priv->dialog_host);
+
+  if (adaptive_preview) {
+    priv->adaptive_preview = adw_adaptive_preview_new ();
+    gtk_window_set_child (GTK_WINDOW (self), priv->adaptive_preview);
+    g_signal_connect_swapped (priv->adaptive_preview, "exit",
+                              G_CALLBACK (adaptive_preview_exit_cb), self);
+    adw_adaptive_preview_set_child (ADW_ADAPTIVE_PREVIEW (priv->adaptive_preview),
+                                    priv->dialog_host);
+  } else {
+    adw_adaptive_preview_set_child (ADW_ADAPTIVE_PREVIEW (priv->adaptive_preview),
+                                    NULL);
+    gtk_window_set_child (GTK_WINDOW (self), priv->dialog_host);
+    priv->adaptive_preview = NULL;
+  }
+
+  g_object_unref (priv->dialog_host);
+
+  g_object_notify_by_pspec (G_OBJECT (self), props[PROP_ADAPTIVE_PREVIEW]);
+}
